/**
  *  \file client/act-planet.cc
  *  \brief Planet Actions
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/actions.h"
#include "client/dlg-cargo.h"
#include "game/global.h"
#include "game/planet.h"
#include "game/ship.h"
#include "ui/commands.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "util/nls.h"
#include "gfx/palette.h"
#include "client/wcost.h"
#include "client/dlg-tech.h"
#include "game/tables.h"
#include "cpluslib/format.h"
#include "game/sim-game.h"
#include "client/scr-sim.h"
#include "game/actions/planetbase.h"
#include "game/actions/basefix.h"
#include "ui/widgets/keytranslator.h"
#include "ui/util.h"
#include "ui/widgets/statictext.h"
#include "ui/widgets/button.h"
#include "game/cargotr.h"
#include "ui/layout.h"
#include "client/widgets/objlist.h"
#include "game/planettransfer.h"
#include "game/shiptransfer.h"
#include "ui/widgets/quit.h"

/** Cargo transfer to/from planet, UI action.
    \param planet    invoking planet
    \param unload    true if 'u' key, false if 'c' key. These have no
                     difference other than the heading of the target
                     selection window. */
void
doPlanetTransfer(GUniverse& univ, GPlanet& planet, bool unload)
{
    GObjectList list;
    for (int i = univ.ty_played_ships.findNextIndex(0); i != 0; i = univ.ty_played_ships.findNextIndex(i)) {
        GShip& ship = univ.getShip(i);
        if (ship.isPlayable(GObject::Playable) && ship.getOwner() == planet.getOwner() && ship.getPos() == planet.getPos()) {
            list.addObject(univ.ty_played_ships, i);
        }
    }
    if (list.size() == 0) {
        messageBox(_("There's none of our ships orbiting this planet."),
                   _("Cargo Transfer"));
        return;
    }
    WObjectList wlist(0, 15, 320, list);
    GObjectReference chosen = wlist.doStandardDialog2(unload ? _("Unload ship...") : _("Transfer cargo to..."), "pcc2:cargo");
    if (chosen.isValid()) {
        doPlanetTransferFor(univ, planet, chosen);
    }
}

/** Planet cargo transfer to ship, UI action.
    \param planet  invoking planet
    \param chosen  target. Must be a ship. */
void
doPlanetTransferFor(GUniverse& univ, GPlanet& planet, GObjectReference target)
{
    GCargoTransfer ct;
    ct.addNewContainer(new GPlanetTransfer(planet));
    ct.addNewContainer(new GShipTransfer(*checked_cast<GShip*>(&*target)));
    doCargoTransfer(ct);
    univ.doScreenUpdates();
}

/** Starbase mission change, UI action. */
void
doBaseChangeMission(GPlanet& planet)
{
    int msn = planet.getBaseMission();
    if (chooseBaseMission(msn, "pcc2:basescreen")) {
        planet.setBaseMission(mp16_t(msn));
    }
}

/** Starbase shipyard order, UI action.
    \param planet the starbase we're talking about.
    \param w      triggering widget, used for binding the menu. */
void
doBaseFixRecycle(GUniverse& univ, GPlanet& planet, UIBaseWidget& w)
{
    GStarbaseFixRecycleAction action(univ, planet);
    int flags = action.getValidBaseActions();
    UIStringList d;
    if ((flags & action.CanFixFlag) != 0)
        d.add(sya_Fix, _("Fix (repair) a ship"));
    if ((flags & action.CanRecycleFlag) != 0)
        d.add(sya_Recycle, _("Recycle a ship"));
    if (planet.getBaseShipyardOrder() == sya_Recycle)
        d.add(sya_None, _("Cancel `recycle' order"));
    if (planet.getBaseShipyardOrder() == sya_Fix)
        d.add(sya_None, _("Cancel `fix' order"));

    if (d.getCount() == 0) {
        messageBox(_("There is no ship at this starbase which could be fixed or recycled."),
                   _("Shipyard"));
        return;
    }

    UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], d) + 20, d);
    if (!lb.doStandardMenu(w))
        return;
    switch (int cmd = lb.getKey()) {
     case sya_Recycle:
     case sya_Fix:
        {
            GObjectList list;
            action.enumerateShipsFor(TShipyardAction(cmd), list);
            list.postprocessUser(true);

            const char* title = (cmd == sya_Recycle ? _("Recycle ship") : _("Fix ship"));
            GObjectReference chosen = WObjectList(0, -1, 300, list).doStandardDialog2(title, "pcc2:basescreen");
            if (chosen.isValid())
                action.setAction(TShipyardAction(cmd), chosen->getId());
        }
        break;
     case sya_None:
        action.setAction(sya_None, 0);
        break;
    }
    univ.doScreenUpdates();
}

void
doPlanetBuildBase(GUniverse& univ, GPlanet& planet)
{
    GMultiTransaction mt;
    GCargoContainer& pay =
        mt.addNewSubtransaction(new GPlanetTransfer(planet));
    GPlanetBuildStarbaseAction& action =
        mt.addNewSubtransaction(new GPlanetBuildStarbaseAction(planet, pay, !planet.isBuildingBase()));

    if (GError e = action.getError()) {
        if (!e.isError(e.eNoResource)) {
            messageBox(e.getUserError(), _("Build Starbase"));
            mt.doCancel();
            return;
        }
    }

    if (planet.isBuildingBase()) {
        bool answer = UIMessageBox(_("You wanted to build a starbase at this planet. "
                                     "Cancel this order?"),
                                   _("Build Starbase")).doYesNoDialog();
        if (answer && !action.getError())
            mt.doCommit();
        else
            mt.doCancel();
        univ.doScreenUpdates();
    } else {
        static const UIKeyTranslator::Item xl[] = {
            { SDLK_RETURN, cm_Enter },
            { ' ',         cm_Enter },
            { SDLK_ESCAPE, cm_Escape }
        };

        UIKeyString y(_("Yes")), n(_("No"));
        UIWindow win(_("Build Starbase"), UIWindow::BLUE_WINDOW, COLOR_GRAY);
        win.setLayout(UIVBoxLayout::instance5);
        ObjectHolder h;
        win.add(h.add(new WCostDisplay(action,
                                       GCargoTypeSet() + el_Tritanium + el_Duranium + el_Molybdenum + el_Money)));
        UIGroup bg(UIHBoxLayout::instance5);
        bg.add(h.add(new UISpacer()));
        if (!action.getError()) {
            win.add(h.add(new UIStaticText(_("Build this starbase?"), UIStaticText::StaticTitle)));
            bg.add(h.add(new UIButton(cm_Enter,  y.getString(), y.getKey())).addStop());
            bg.add(h.add(new UIButton(cm_Escape, n.getString(), n.getKey())).addStop());
        } else {
            win.add(h.add(new UIStaticText(_("You can't build a starbase here."), UIStaticText::StaticTitle)));
            bg.add(h.add(new UIButton(cm_Escape, _("OK"), SDLK_RETURN)).addStop());
        }
        bg.add(h.add(new UISpacer()));
        win.add(bg);
        win.add(h.add(new UIKeyTranslator(xl, countof(xl)))
                .addStop(y.getKey(), cm_Enter)
                .addStop(n.getKey(), cm_Escape));
        win.add(h.add(new UIQuit(cm_Escape)));
        win.pack();
        win.setState(win.st_Modal, true);

        ui_root->centerWidget(win);
        ui_root->add(win);
        int rv = ui_root->run();
        ui_root->close(win);
        if (rv == cm_Enter && !action.getError())
            mt.doCommit();
        else
            mt.doCancel();
        univ.doScreenUpdates();
    }
}

void
doPlanetAddToSim(GPlanet& planet)
{
    if (!WSimScreen::sim_state.hasPlanet())
        WSimScreen::sim_state.addPlanet();

    GSimulatorRealGameInterface::copyPlanet(WSimScreen::sim_state.getPlanet(), planet);
    WSimScreen::sim_state.doUpdates();
}

bool
chooseBaseMission(int& value, const char* help)
{
    UIStringList data;
    for (int i = 0; i <= NUM_BASE_MISSIONS; ++i)
        data.add(i, format("%d - %s", i, getBaseMissionName(i)));

    UIStandardListbox lb(0, 7, 300, data);
    lb.scrollToKey(value);
    if (lb.doStandardListbox(_("Starbase Mission"), help)) {
        value = lb.getKey();
        return true;
    } else {
        return false;
    }
}
