/**
  *  \file client/dialogs/allydlg.cc
  *  \brief Alliance Dialog
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/dialogs/allydlg.h"
#include "client/widgets/allylevelgrid.h"
#include "client/widgets/allylist.h"
#include "client/widgets/helpwidget.h"
#include "game/alliance.h"
#include "game/global.h"
#include "game/team.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/msgbox.h"
#include "ui/uiglobal.h"
#include "ui/widgets/button.h"
#include "ui/widgets/focusiterator.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/statictext.h"
#include "ui/window.h"
#include "util/nls.h"
#include "game/pref.h"
#include "ui/widgets/quit.h"

namespace {
    /** Alliance editor window.
        This window lets the user edit their alliances.

        It works on a copy of a GAlliances object.
        Upon OK, the changes can be copied back into a live object using copyTo().
        This will generate the command messages. */
    class WAllyWindow : public UIWindow {
     public:
        WAllyWindow(const GAlliances& liveAllies);
        void copyTo(GAlliances& liveAllies);

     private:
        void init();

        ObjectHolder h;
        GAlliances editAllies;

        WAllyList* pAllyList;
        WAllyLevelGrid* pAllyGrid;

        /** Status updater.
            This is registered as a change handler to the local GAlliances copy
            and manages redraw. */
        class Updater : public GAllianceHandler {
         public:
            Updater(WAllyWindow& window);
            void postprocess(GAlliances&);
            void handleChanges(const GAlliances&);
         private:
            WAllyWindow& window;
        };
        friend class Updater;
    };
}

/** Constructor. */
WAllyWindow::WAllyWindow(const GAlliances& liveAllies)
    : UIWindow(_("Edit Alliances"), UIWindow::BLUE_WINDOW, COLOR_GRAY),
      editAllies(liveAllies),
      pAllyList(0),
      pAllyGrid(0)
{
    init();
}

void
WAllyWindow::copyTo(GAlliances& liveAllies)
{
    liveAllies.copyFrom(editAllies);
}

void
WAllyWindow::init()
{
    // VBox
    //   HBox
    //     VBox
    //       Static "Alliances:"
    //       WAllyList
    //       Static help text
    //       UISpacer
    //     VBox
    //       Static "Status:"
    //       WAllyLevelGrid
    //       UISpacer
    //   HBox
    //     "OK"
    //     "Cancel"
    //    ["Teams"]
    //     UISpacer
    //     "Help"

    pAllyList = &h.add(new WAllyList(1000, getPlayerId(), editAllies));
    pAllyGrid = &h.add(new WAllyLevelGrid(1001, editAllies));

    setLayout(UIVBoxLayout::instance5);

    UIGroup& g1 = h.add(new UIGroup(UIHBoxLayout::instance5));
    UIGroup& g11 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g11.add(h.add(new UIStaticText(_("Alliances:"), UIStaticText::StaticTitle)));
    g11.add(UIFrameGroup::wrap(h, *pAllyList));
    g11.add(h.add(new UISpacer()));
    g1.add(g11);
    
    UIGroup& g12 = h.add(new UIGroup(UIVBoxLayout::instance5));
    g12.add(h.add(new UIStaticText(_("Status:"), UIStaticText::StaticTitle)));
    g12.add(*pAllyGrid);
    g12.add(h.add(new UISpacer()));
    g1.add(g12);
    add(g1);

    UIGroup& g2 = h.add(new UIGroup(UIHBoxLayout::instance5));
    g2.add(h.add(new UIButton(1, _("OK"), SDLK_RETURN)).addStop());
    g2.add(h.add(new UIButton(0, _("Cancel"), SDLK_ESCAPE)).addStop());
    g2.add(h.add(new UISpacer()));
    g2.add(h.add(new UIButton(_("Help"), 'h')));
    add(g2);
    add(h.add(new UIFocusIterator(1000, 1001, UIFocusIterator::fi_Tab | UIFocusIterator::fi_Horiz)));
    add(h.add(new WHelpWidget("pcc2:allies")));
    add(h.add(new UIQuit(0)));

    pack();

    pAllyList->setFocusedWidget();
    pAllyGrid->setPlayer(pAllyList->getPosition());
    pAllyList->sig_player_change.add(pAllyGrid, &WAllyLevelGrid::setPlayer);

    // This is kind-of hacky: add an observer to get notified when someone edits the editAllies.
    editAllies.addNewHandler(new Updater(*this));
}


inline
WAllyWindow::Updater::Updater(WAllyWindow& window)
    : window(window)
{ }

void
WAllyWindow::Updater::postprocess(GAlliances&)
{ }

void
WAllyWindow::Updater::handleChanges(const GAlliances&)
{
    window.pAllyList->drawWidget();
    window.pAllyGrid->drawWidget();
}

/** Alliance dialog.
    Allows the user to edit their alliances. */
void
doAllianceDialog()
{
    // liveAllies is not necessarily in sync with command messages; update it.
    GAlliances& liveAllies = getCurrentTurn().getAlliances(getPlayerId());
    liveAllies.postprocess();

    // Do we actually allow alliances?
    if (liveAllies.getLevels().size() == 0) {
        messageBox(_("Your host does not support alliances, or PCC2 does not know how to configure them."), _("Edit Alliances"));
        return;
    }

    // Build the dialog.
    WAllyWindow window(liveAllies);
    ui_root->centerWidget(window);
    if (ui_root->runModal(window)) {
        // User confirmed; write back. This will update command messages.
        window.copyTo(liveAllies);

        // Update teams if configured
        syncTeamsFromAlliances(liveAllies);
    }
}
