/**
  *  \file client/dialogs/buildqueuesummary.cc
  *  \brief Build Queue Summary Dialog
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "client/dialogs/buildqueuesummary.h"
#include "client/widgets/helpwidget.h"
#include "cpluslib/format.h"
#include "cpluslib/holder.h"
#include "game/sel.h"
#include "gfx/palette.h"
#include "ui/layout.h"
#include "ui/listbox.h"
#include "ui/util.h"
#include "ui/widgets/button.h"
#include "ui/widgets/framegroup.h"
#include "ui/widgets/quit.h"
#include "util/nls.h"

namespace {

    /*
     *  Matching items
     */

    // Match mode
    enum Match {
        MatchNone,
        MatchAll,
        MatchAction,
        MatchHull
    };
    typedef std::pair<Match, int> Match_t;

    /* Check for match of one build queue entry */
    bool matchInfo(const Match_t& m, const GChangeBuildQueue::Info& in)
    {
        bool ok = false;
        switch (m.first) {
         case MatchNone:                                      break;
         case MatchAll:    ok = true;                         break;
         case MatchAction: ok = (m.second == int(in.action)); break;
         case MatchHull:   ok = (m.second == in.hullNr);      break;
        }
        return ok;
    }


    /*
     *  Hull summary
     */

    // Summary entry for one hull
    struct HullEntry {
        int numBuild, numPlan, numClone;
        const GChangeBuildQueue::Info* pSpecimen;

        HullEntry()
            : numBuild(0), numPlan(0), numClone(0), pSpecimen(0)
            { }
    };
    typedef std::map<int, HullEntry> HullEntries_t;

    /* Account for one build queue entry */
    void addInfo(HullEntry& out, const GChangeBuildQueue::Info& in)
    {
        switch (in.action) {
         case GChangeBuildQueue::BuildShip: ++out.numBuild; break;
         case GChangeBuildQueue::CloneShip: ++out.numClone; break;
         case GChangeBuildQueue::PlanShip:  ++out.numPlan;  break;
        }
        out.pSpecimen = &in;
    }


    /*
     *  Comparison
     */

    struct CompareHullEntriesByName {
        bool operator()(const HullEntry* a, const HullEntry* b) const
            {
                return a != 0 && a->pSpecimen != 0
                    && b != 0 && b->pSpecimen != 0
                    && (a->pSpecimen->hullName < b->pSpecimen->hullName);
            }
    };


    /*
     *  Build Queue Summary Listbox
     *
     *  We use a RichListbox implementation, but add some extra functions to set it up.
     */

    class BuildQueueSummaryListbox : public UIListbox {
     public:
        BuildQueueSummaryListbox();

        void addSummaryItem(Match_t m, string_t label, string_t info);
        void addSummaryHeading(string_t label);
        void addOrderItem(Match_t m, string_t label, int num);
        void addHullItem(const HullEntry& entry);

        Match_t getCurrentMatcher() const;

        // UIListbox:
        virtual bool isAccessible(int pos);
        virtual void drawPart(GfxCanvas& can, int from, int to);

     private:
        struct Entry {
            Match_t match;
            string_t label;
            string_t info;
            Entry(Match_t match, string_t label, string_t info)
                : match(match), label(label), info(info)
                { }
        };
        std::vector<Entry> m_entries;
    };


    /*
     *  Dialog
     */

    class BuildQueueSummaryDialog {
     public:
        BuildQueueSummaryDialog(BuildQueueSummaryListbox& list, const GChangeBuildQueue::Infos_t& infos)
            : m_list(list),
              m_infos(infos),
              m_markButton(0, _("Mark..."), 'm')
            {
                m_markButton.sig_fired.add(this, &BuildQueueSummaryDialog::onMark);
            }

        void run()
            {
                // Window [VBox]
                //   FrameGroup > ScrollbarContainer > List
                //   HBox
                //      Button "Close"
                //      Button "Mark..."
                //      Spacer
                //      Button "Help"
                ObjectHolder h;
                UIWindow& win = h.add(new UIWindow(_("Build Order Summary"), UIWindow::BLUE_WINDOW, COLOR_GRAY));
                win.setLayout(UIVBoxLayout::instance5);
                win.add(UIFrameGroup::wrap(h, m_list));

                UIGroup& buttonGroup = h.add(new UIGroup(UIHBoxLayout::instance5));
                UIButton& btnClose = h.add(new UIButton(0, _("Close"), SDLK_ESCAPE));
                UIButton& btnHelp  = h.add(new UIButton(_("Help"),  'h'));
                buttonGroup.add(btnClose);
                buttonGroup.add(m_markButton);
                buttonGroup.add(h.add(new UISpacer()));
                buttonGroup.add(btnHelp);
                win.add(buttonGroup);
                win.add(h.add(new WHelpWidget("pcc2:queuemanager")));
                win.add(h.add(new UIQuit(0)));
                win.pack();

                // Limit size
                int rw = ui_root->getExtent().w;
                int rh = ui_root->getExtent().h;
                int nw = std::min(rw * 9/10, int(win.getExtent().w));
                int nh = std::min(rh * 9/10, int(win.getExtent().h));
                // m_list.setPreferredHeight(rh * 7/10);
                win.setExtent(GfxRect(0, 0, nw, nh));

                btnClose.addStop();
                // btnHelp.dispatchKeyTo(helpWidget);

                ui_root->centerWidget(win);
                ui_root->runModal(win);
            }

        void onMark()
            {
                // Menu
                enum { Mark, MarkOnly, Unmark };

                StringList list;
                list.add(Mark,     _("Mark these starbases"));
                list.add(MarkOnly, _("Mark only these starbases"));
                list.add(Unmark,   _("Unmark these starbases"));

                UIStandardListbox lb(0, 1, getTextWidth(*fonts[FONT_NORMAL], list) + 20, list);
                if (!lb.doStandardMenu(m_markButton)) {
                    return;
                }

                // Build list
                switch (lb.getKey()) {
                 case Mark:
                    markMatch(true);
                    break;
                 case MarkOnly:
                    GSelection().copyTo(getDisplayedTurn().getCurrentUniverse());
                    markMatch(true);
                    break;
                 case Unmark:
                    markMatch(false);
                    break;
                }
            }

        void markMatch(bool value)
            {
                const Match_t m = m_list.getCurrentMatcher();
                GUniverse& univ = getDisplayedTurn().getCurrentUniverse();
                for (size_t i = 0, n = m_infos.size(); i < n; ++i) {
                    if (matchInfo(m, m_infos[i])) {
                        const int id = m_infos[i].planetId;
                        if (univ.ty_played_planets.isValidIndex(id)) {
                            univ.ty_played_planets.getObjectByIndex(id).setIsMarked(value);
                        }
                    }
                }
                univ.doScreenUpdates();
            }

     private:
        BuildQueueSummaryListbox& m_list;
        const GChangeBuildQueue::Infos_t& m_infos;
        UIButton m_markButton;
    };
}


/*
 *  BuildQueueSummaryListbox implementation
 */

BuildQueueSummaryListbox::BuildQueueSummaryListbox()
    : UIListbox(0, 20, 400, 1)
{ }

void
BuildQueueSummaryListbox::addSummaryItem(Match_t m, string_t label, string_t info)
{
    m_entries.push_back(Entry(m, label, info));
    setSize(m_entries.size());
}

void
BuildQueueSummaryListbox::addSummaryHeading(string_t label)
{
    m_entries.push_back(Entry(Match_t(MatchNone, 0), label, string_t()));
    setSize(m_entries.size());
}

void
BuildQueueSummaryListbox::addOrderItem(Match_t m, string_t label, int num)
{
    if (num != 0) {
        addSummaryItem(m, label, format("(%d)", num));
    }
}

void
BuildQueueSummaryListbox::addHullItem(const HullEntry& entry)
{
    if (entry.pSpecimen != 0) { // should never fail
        string_t extra;
        if (entry.numBuild != 0) {
            extra += format(_("%d\xC3\x97 build"), entry.numBuild);
        }
        if (entry.numClone != 0) {
            if (!extra.empty()) extra += ", ";
            extra += format(_("%d\xC3\x97 clone"), entry.numClone);
        }
        if (entry.numPlan != 0) {
            if (!extra.empty()) extra += ", ";
            extra += format(_("%d\xC3\x97 plan"), entry.numPlan);
        }
        addSummaryItem(Match_t(MatchHull, entry.pSpecimen->hullNr), entry.pSpecimen->hullName, "(" + extra + ")");
    }
}

Match_t
BuildQueueSummaryListbox::getCurrentMatcher() const
{
    int n = getCurrent();
    return ((n >= 0 && n < int(m_entries.size()))
            ? m_entries[n].match
            : Match_t());
}

bool
BuildQueueSummaryListbox::isAccessible(int pos)
{
    return pos >= 0
        && pos < int(m_entries.size())
        && m_entries[pos].match.first != MatchNone;
}

void
BuildQueueSummaryListbox::drawPart(GfxCanvas& can, int from, int to)
{
    GfxContext ctx;
    ctx.useCanvas(can);
    for (int i = from; i < to; ++i) {
        GfxRect area;
        prepareEntry(ctx, ColorStyle, i, &area);
        area.consumeX(5);

        int index = i + getTop();
        if (index >= 0 && index < int(m_entries.size())) {
            const Entry& e = m_entries[index];
            if (e.match.first == MatchNone) {
                // Heading
                ctx.useFont(*fonts[FONT_NORMAL_BOLD]);
                ctx.setColor(UIColor::tc_Static);
                outTextF(ctx, area, e.label);
            } else {
                // Category
                ctx.useFont(*fonts[FONT_NORMAL]);
                ctx.setColor(UIColor::tc_Static);

                string_t label = "  " + e.label + " ";
                outTextF(ctx, area.splitX(fonts[FONT_NORMAL]->getTextWidth(label)), label);
                ctx.setColor(UIColor::tc_Faded);
                outTextF(ctx, area, e.info);
            }
        }
    }
}


/*
 *  Entry Point
 */

void
doBuildQueueSummaryDialog(const GChangeBuildQueue::Infos_t& infos)
{
    // Count everything
    HullEntries_t byHull;
    HullEntry summary;
    for (size_t i = 0; i < infos.size(); ++i) {
        addInfo(byHull[infos[i].hullNr], infos[i]);
        addInfo(summary, infos[i]);
    }

    // Sort by hull name
    std::vector<const HullEntry*> entries;
    for (HullEntries_t::const_iterator it = byHull.begin(); it != byHull.end(); ++it) {
        entries.push_back(&it->second);
    }
    std::sort(entries.begin(), entries.end(), CompareHullEntriesByName());

    // Build list
    BuildQueueSummaryListbox box;
    box.addSummaryHeading(_("Orders"));
    box.addOrderItem(Match_t(MatchAll, 0), _("All"), summary.numPlan + summary.numClone + summary.numBuild);
    box.addOrderItem(Match_t(MatchAction, GChangeBuildQueue::BuildShip), _("Build"), summary.numBuild);
    box.addOrderItem(Match_t(MatchAction, GChangeBuildQueue::CloneShip), _("Clone"), summary.numClone);
    box.addOrderItem(Match_t(MatchAction, GChangeBuildQueue::PlanShip),  _("Plan"),  summary.numPlan);
    box.addSummaryHeading(_("Ship Types"));
    for (size_t i = 0, n = entries.size(); i < n; ++i) {
        box.addHullItem(*entries[i]);
    }
    box.scrollTo(0, 1);

    // Dialog
    BuildQueueSummaryDialog(box, infos).run();
}
