/**
  *  \file client/screens/basescreen.cc
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "client/screens/basescreen.h"
#include "client/cscreen.h"
#include "ui/msgbox.h"
#include "game/objsel.h"
#include "client/tiles/baseoverview.h"
#include "client/dialogs/objselect.h"
#include "int/keymap.h"
#include "game/global.h"
#include "util/nls.h"
#include "ui/window.h"
#include "client/wscanchart.h"
#include "client/tiles/baseorder.h"
#include "ui/uiglobal.h"
#include "client/actions.h"
#include "int/if/planetif.h"

namespace {
    class WBaseScreen : public WControlScreen {
     public:
        WBaseScreen();
        ~WBaseScreen();
        bool handleCommand(const UICommand& cmd, UIBaseWidget& sender);
        void update();
        virtual void enumContexts(ptr_vector<IntContext>& ctx);
    };
}

int
chooseBase()
{
    if (current_base_type.isEmpty()) {
        messageBox(_("You do not have any starbases. "
                     "To build starbases, use [F8] on a planet."),
                   _("Starbases"));
        return 0;
    } else if (current_base_type.isUnit()) {
        return current_base_selection.getCurrentIndex();
    } else {
        GSimpleObjectSelection select;
        select.initFrom(current_base_selection);
        WBaseOverviewTile bot(select);
        WObjectSelectionDialog dlg(_("Select Starbase"), "pcc2:unitsel", select);
        dlg.addWidget(bot);
        dlg.pack();
        ui_root->centerWidget(dlg);
        if (ui_root->runModal(dlg) == cm_Enter) {
            return select.getCurrentIndex();
        } else {
            return 0;
        }
    }
}

/****************************** WBaseScreen ******************************/

WBaseScreen::WBaseScreen()
    : WControlScreen(current_base_selection, WScreenId::BaseScreen)
{
    createTileLayout(*this, "BASESCREEN", current_base_selection);
    addChart(holder.add(new WScannerChartWidget()));
    setKeymap(IntKeymap::getKeymapByName("BASESCREEN"));
}

WBaseScreen::~WBaseScreen()
{ }

bool
WBaseScreen::handleCommand(const UICommand& cmd, UIBaseWidget& sender)
{
    if (cmd.code == cm_Base_Recycle) {
        int pid         = getObjectSelection().getCurrentIndex();
        GUniverse& univ = *getObjectSelection().getCurrentUniverse();
        GPlanet& me     = univ.getPlanet(pid);
        doBaseFixRecycle(univ, me, sender);
        return true;
    }
    return WControlScreen::handleCommand(cmd, sender);
}

void
WBaseScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(getObjectSelection().getCurrentObject())) {
        getScanner().getChart()->setCenterPosition(mo->getPos());
        getScanner().getChart()->setScannerOrigin(mo->getPos());
        getScanner().getChart()->setScannerTarget(mo->getPos());
    }
}

void
WBaseScreen::enumContexts(ptr_vector<IntContext>& ctx)
{
    ctx.push_back(new IntPlanetContext(getObjectSelection().getCurrentIndex()));
}

void
doBaseScreen()
{
    return WBaseScreen().runScreen();
}
