/**
  *  \file client/screens/shiptaskscreen.cc
  *  \brief Ship Task Screen
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cmath>
#include "client/screens/shiptaskscreen.h"
#include "client/wscanchart.h"
#include "cpluslib/format.h"
#include "game/fcode.h"
#include "game/global.h"
#include "game/host.h"
#include "game/obj.h"
#include "game/ship.h"
#include "gfx/palette.h"
#include "int/exec.h"
#include "int/if/shipif.h"
#include "int/intglobal.h"
#include "int/process.h"
#include "util/nls.h"
#include "client/tasksupport.h"
#include "ui/uiglobal.h"
#include "client/screens/shipscreen.h"

namespace {
    class WShipTaskScannerChartWidget : public WScannerChartWidget {
        SignalConnection conn_change;
        WShipAutoTaskSelection& selection;
     public:
        WShipTaskScannerChartWidget(WShipAutoTaskSelection& selection);
        void lockQueryLocation(bool& hyperjumping, int& current_gravity_pid);
        int32_t lockQueryDistance(GPoint pt);
        void drawPre(GfxCanvas& can);
        void drawPost(GfxCanvas& can);
        void onDblClick();
    };

    class WShipAutoTaskScreen : public WAutoTaskScreen {
        WShipTaskScannerChartWidget* scanner;
     public:
        WShipAutoTaskScreen(WShipAutoTaskSelection& selection);
        void update();
        bool handleEvent(const UIEvent& event, bool second_pass);
        virtual void enumContexts(ptr_vector<IntContext>& ctx);
    };
}

// Config options. FIXME: make persistent
static bool predictToEnd = false;
static bool showDistances = true;

/*************************** WShipAutoTaskSelection ***************************/

WShipAutoTaskSelection::WShipAutoTaskSelection(GObjectSelection& base)
    : WAutoTaskObjectSelection(base)
{
    sig_task_changed.add(this, &WShipAutoTaskSelection::onTaskChange);
}

IntExecutionContext*
WShipAutoTaskSelection::getTask(bool create)
{
    return getAutoTaskForObject(*getCurrentObject(), IntExecutionContext::pkShipTask, create);
}

void
WShipAutoTaskSelection::onTaskChange()
{
    predictor.reset(new IntShipPredictor(getCurrentIndex()));
    if (isPredictToEnd() || getCursor() < getEditor().getPC()) {
        predictor->predictTask(getEditor());
    } else {
        predictor->predictTask(getEditor(), getCursor());
    }
}

void
WShipAutoTaskSelection::insertWaypointCommand(const char* cmd, GPoint pt, bool setWarp, bool ignoreDuplicate)
{
    // First figure out current state
    IntShipPredictor pred(getCurrentIndex());
    pred.predictTask(getEditor(), getCursor());
    if (ignoreDuplicate && pred.getPosition() == pt) {
        return;
    }

    // Set speed if desired
    GShip* sh = dynamic_cast<GShip*>(getCurrentObject());
    if (setWarp && sh != 0) {
        // Do it
        int32_t dist2 = distanceSquared(pred.getPosition(), pt);
        if (sh->canDoSpecial(hf_Hyperdrive) && host.isExactHyperjumpDistance2(dist2)) {
            /* Looks like a hyperjump, so make one. This code is not in the regular
               auto-warp function, but is's very convenient for planning double-jumps. */
            int speed = std::max(2, getFleetEngineSpeed(getDisplayedTurn().getCurrentUniverse(), sh->getId()));
            if (pred.getSpeed() < speed) {
                insert(format("SetSpeed %d", speed));
            }
            insert(format("SetFCode \"HYP\"   %% %s", _("hyperjump")));
        } else {
            /* Not a hyperjump */
            if (sh->canDoSpecial(hf_Hyperdrive)
                && pred.getSpeed() > 0
                && pred.getFCode() == "HYP"
                && !host.isExactHyperjumpDistance2(dist2))
            {
                insert(format("SetFCode %s   %% %s",
                              quoteString(GFCode::generateRandomFCode()),
                              _("cancel hyperjump")));
            }

            /* Optimize speed */
            if (pred.getPosition() != pt) {
                int n = computeOptimumWarp(getDisplayedTurn().getCurrentUniverse(), getCurrentIndex(), pred.getPosition(), pt);
                if (n != 0 && n != pred.getSpeed()) {
                    insert(format("SetSpeed %d", n));
                }
            }
        }
    }

    // Now the waypoint command
    string_t command = format("%s %d, %d", cmd, pt.x, pt.y);
    string_t comment = getDisplayedTurn().getCurrentUniverse().getLocationName(pt, locs_WW | locs_NoSpace);
    if (comment.size()) {
        command += "   % ";
        command += comment;
    }
    insert(command);
}

bool
WShipAutoTaskSelection::isPredictToEnd()
{
    return predictToEnd;
}

void
WShipAutoTaskSelection::setPredictToEnd(bool value)
{
    if (value != predictToEnd) {
        predictToEnd = value;
        sig_task_changed.raise(false);
    }
}

bool
WShipAutoTaskSelection::isShowDistances()
{
    return showDistances;
}

void
WShipAutoTaskSelection::setShowDistances(bool value)
{
    if (value != showDistances) {
        showDistances = value;
        sig_task_changed.raise(false);
    }
}

const IntShipPredictor*
WShipAutoTaskSelection::getPredictor() const
{
    return predictor.get();
}

/********************** WShipTaskScannerChartWidget **********************/

WShipTaskScannerChartWidget::WShipTaskScannerChartWidget(WShipAutoTaskSelection& selection)
    : conn_change(selection.sig_task_changed.add(static_cast<UIBaseWidget*>(this),
                                            (void (UIBaseWidget::*)()) &UIBaseWidget::drawWidget)),
      selection(selection)
{
    sig_dblclick.add(this, &WShipTaskScannerChartWidget::onDblClick);
}

void
WShipTaskScannerChartWidget::lockQueryLocation(bool& hyperjumping, int& current_gravity_pid)
{
    const GShip* sh = dynamic_cast<GShip*>(selection.getCurrentObject());
    const IntShipPredictor* pred = selection.getPredictor();
    if (sh != 0 && pred != 0) {
        GPoint pt;
        if (pred->getNumPositions() > 0) {
            pt = pred->getPosition(pred->getNumPositions()-1);
        } else {
            pt = pred->getPosition();
        }

        hyperjumping = pred->isHyperdriving();
        current_gravity_pid = getDisplayedTurn().getCurrentUniverse().getPlanetAt(pt, true);
    } else {
        WScannerChartWidget::lockQueryLocation(hyperjumping, current_gravity_pid);
    }
}

int32_t
WShipTaskScannerChartWidget::lockQueryDistance(GPoint pt)
{
    const GShip* sh = dynamic_cast<GShip*>(selection.getCurrentObject());
    const IntShipPredictor* pred = selection.getPredictor();
    if (pred != 0 && sh != 0) {
        int32_t dist2;
        GPoint pos;
        if (pred->getNumPositions() > 0) {
            pos = pred->getPosition(pred->getNumPositions()-1);
        } else {
            pos = pred->getPosition();
        }
        dist2 = distanceSquared(pt, pos);

        GShipTurnPredictor pred(getDisplayedTurn().getCurrentUniverse(), sh->getId());
        pred.setPosition(pos);
        pred.setSpeed(getEngine(sh->getEngineType()).getMaxEfficientWarp());
        pred.setWaypoint(pt);
        pred.computeMovement();
        int t = pred.getNumTurns();

        return combineTimeAndDistance2(dist2, t);
    } else {
        return WScannerChartWidget::lockQueryDistance(pt);
    }
}

/** Draw pre-pass. This draws the movement segments. */
void
WShipTaskScannerChartWidget::drawPre(GfxCanvas& can)
{
    WScannerChartWidget::drawPre(can);

    GfxContext ctx;
    ctx.useCanvas(can);
    ctx.useColorScheme(GfxStandardColorScheme::instance);

    GChartViewport vp = getViewport();

    const IntShipPredictor* pred = selection.getPredictor();
    if (pred) {
        GPoint pt = getScannerOrigin();
        for (int i = 0; i < pred->getNumPositions(); ++i) {
            int pix;
            if (i < pred->getNumFuelPositions()) {
                ctx.setColor(COLOR_DARKYELLOWSCALE+6);
                ctx.setLinePattern(SOLID_LINE);
                pix = COLOR_YELLOW;
            } else {
                ctx.setColor(COLOR_DARK);
                ctx.setLinePattern(DOTTED_LINE);
                pix = COLOR_GRAY;
            }

            GPoint alias = pred->getPosition(i).getSimpleNearestAlias(pt);
            GfxPoint npt = vp.scaleXY(alias);
            drawLine(ctx, vp.scaleXY(pt), npt);

            ctx.setColor(pix);
            drawPixel(ctx, npt.movedBy(1, 0));
            drawPixel(ctx, npt.movedBy(-1, 0));
            drawPixel(ctx, npt.movedBy(0, 1));
            drawPixel(ctx, npt.movedBy(0, -1));
            pt = alias;
        }
    }
}

/** Draw post-pass. Distances are drawn here, to be readable atop
    other starchart items like ion storms. */
void
WShipTaskScannerChartWidget::drawPost(GfxCanvas& can)
{
    WScannerChartWidget::drawPost(can);

    const IntShipPredictor* pred = selection.getPredictor();

    if (pred && selection.isShowDistances()) {
        GfxContext ctx;
        ctx.useCanvas(can);
        ctx.useColorScheme(GfxStandardColorScheme::instance);
        ctx.useFont(*fonts[FONT_SMALL]);

        GChartViewport vp = getViewport();

        GPoint pt = getScannerOrigin();
        for (int i = 0; i < pred->getNumPositions(); ++i) {
            if (i < pred->getNumFuelPositions()) {
                ctx.setColor(COLOR_DARKYELLOW);
            } else {
                ctx.setColor(COLOR_GRAY);
            }

            /* Draw only segments of more than 10 ly to avoid cluttering
               up display too badly. */
            uint32_t d2 = distanceSquared(pt, pred->getPosition(i));
            GPoint alias = pred->getPosition(i).getSimpleNearestAlias(pt);
            if (d2 > 100) {
                /* Place distances atop each segment. If the segment is
                   falling, draw to the right [\''], if the segment is
                   rising, draw to the left [''/]. */
                GfxPoint npt = vp.scaleXY(alias);
                GfxPoint opt = vp.scaleXY(pt);
                if ((npt.y < opt.y) != (npt.x < opt.x)) {
                    ctx.setTextAlign(2, 2);
                } else {
                    ctx.setTextAlign(0, 2);
                }
                outText(ctx, (npt.x + opt.x)/2, (npt.y + opt.y)/2,
                        format("%.0f ly", std::sqrt(double(d2))));
            }
            pt = alias;
        }
    }
}

void
WShipTaskScannerChartWidget::onDblClick()
{
    selection.insertWaypointCommand("MoveTo", getScannerTarget(), true, true);
}

/************************** WShipAutoTaskScreen **************************/

WShipAutoTaskScreen::WShipAutoTaskScreen(WShipAutoTaskSelection& selection)
    : WAutoTaskScreen(selection, WScreenId::ShipTaskScreen)
{
    createTileLayout(*this, "SHIPTASKSCREEN", selection);
    scanner = &holder.add(new WShipTaskScannerChartWidget(selection));
    addChart(*scanner);
    setKeymap(IntKeymap::getKeymapByName("SHIPTASKSCREEN"));
}

void
WShipAutoTaskScreen::enumContexts(ptr_vector<IntContext>& ctx)
{
    ctx.push_back(new IntShipContext(selection.getCurrentIndex()));
}

void
WShipAutoTaskScreen::update()
{
    if (GMapObject* mo = dynamic_cast<GMapObject*>(selection.getCurrentObject())) {
        WScannerChartWidget& chart = *getScanner().getChart();
        chart.setCenterPosition(mo->getPos());
        chart.setScannerOrigin(mo->getPos());
        chart.setScannerTarget(mo->getPos());
    }
}

bool
WShipAutoTaskScreen::handleEvent(const UIEvent& event, bool second_pass)
{
    if (event.type == ev_Key) {
        switch (event.key.code) {
            // a, g => WShipAutoTaskPredictionTile
         case ss_Ctrl + 'w':
            scanner->onDblClick();
            return true;
        }
    }
    return WAutoTaskScreen::handleEvent(event, second_pass);
}

void
doShipAutoTaskScreen()
{
    WShipAutoTaskSelection at(current_ship_selection);
    WShipAutoTaskScreen screen(at);
    at.init();
    screen.runScreen();
    at.done();
}
