/**
  *  \file game/hullf-list.cc
  *  \brief Hull Function Lists
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <algorithm>
#include "game/hullfunc.h"
#include "cpluslib/assert.h"

/** Default constructor. Makes a blank list. */
GHullFunctionList::GHullFunctionList()
{ }

/** Destructor. */
GHullFunctionList::~GHullFunctionList()
{ }

/** Append new item at end. */
void
GHullFunctionList::add(const GHullFunction& f)
{
    data.push_back(f);
}

/** Get number of items in list. */
GHullFunctionList::container_t::size_type
GHullFunctionList::size() const
{
    return data.size();
}

/** Get iterator to first item. */
GHullFunctionList::iterator
GHullFunctionList::begin() const
{
    return data.begin();
}

/** Get iterator to one-past-end. */
GHullFunctionList::iterator
GHullFunctionList::end() const
{
    return data.end();
}

/** Clear list. */
void
GHullFunctionList::clear()
{
    data.clear();
}

/** Indexed access.
    \param i Index, [0,size()) */
const GHullFunction&
GHullFunctionList::operator[](container_t::size_type i) const
{
    ASSERT(i < data.size());
    return data[i];
}

static int
firstLevel(GExpLevelSet s)
{
    int f = 0;
    while (f <= NUM_EXPERIENCE_LEVELS && !s.contains(f))
        ++f;
    return f;
}

namespace {
    /** Sort predicate for GHullFunctionList::simplify. Sort by
        function, then levels, then kind, then players. */
    struct SortByFunctionAndLevel {
        bool operator()(const GHullFunction& lhs, const GHullFunction& rhs) const {
            if (lhs.getBasicFunctionId() != rhs.getBasicFunctionId())
                return lhs.getBasicFunctionId() < rhs.getBasicFunctionId();
            if (lhs.getLevels().toInteger() != rhs.getLevels().toInteger())
                return lhs.getLevels().toInteger() < rhs.getLevels().toInteger();
            if (lhs.getKind() != rhs.getKind())
                return lhs.getKind() < rhs.getKind();
            return lhs.getPlayers().toInteger() < rhs.getPlayers().toInteger();
        }
    };

    /** Sort predicate for players. */
    struct SortByPlayer {
        GPlayerSet forPlayer;
        SortByPlayer(GPlayerSet forPlayer)
            : forPlayer(forPlayer)
            { }
        bool operator()(const GHullFunction& lhs, const GHullFunction& rhs) const {
            /* Racial abilities are boring and go last */
            bool lrace = lhs.getKind() == GHullFunction::AssignedToRace;
            bool rrace = rhs.getKind() == GHullFunction::AssignedToRace;
            if (lrace != rrace)
                return lrace < rrace;

            /* Functions available to forPlayer go first */
            bool lsee = lhs.getPlayers().containsAnyOf(forPlayer);
            bool rsee = rhs.getPlayers().containsAnyOf(forPlayer);
            if (lsee != rsee)
                return lsee > rsee;

            /* Next, sort by level */
            int lfirst = firstLevel(lhs.getLevels());
            int rfirst = firstLevel(rhs.getLevels());
            if (lfirst != rfirst)
                return lfirst < rfirst;

            /* Then, functions */
            if (lhs.getBasicFunctionId() != rhs.getBasicFunctionId())
                return lhs.getBasicFunctionId() < rhs.getBasicFunctionId();

            /* Finally, kind and players */
            if (lhs.getKind() != rhs.getKind())
                return lhs.getKind() < rhs.getKind();
            return lhs.getPlayers().toInteger() < rhs.getPlayers().toInteger();
        }
    };
}

void
GHullFunctionList::simplify()
{
    container_t::size_type in, out;

    /* Early exit. */
    // The below sort() line fails when size() is zero on BCC, why?
    if (data.size() == 0)
        return;

    /* Pass 1: combine identical functions, and wield out empty assignments. */
    std::sort(data.begin(), data.end(), SortByFunctionAndLevel());
    in = out = 0;
    while (in < data.size()) {
        GHullFunction func = data[in++];
        while (in < data.size() && data[in].isSame(func) && data[in].getKind() == func.getKind()) {
            func.setPlayers(func.getPlayers() | data[in].getPlayers());
            ++in;
        }
        if (func.getPlayers().nonempty())
            data[out++] = func;
    }
    data.resize(out);

    /* Pass 2: remove racial abilities that are a strict subset of
       hull functions, and vice versa */
    /* FIXME: what do we do with hull vs. ship functions? */
    CTASSERT(GHullFunction::AssignedToHull+1 == GHullFunction::AssignedToRace);
    in = out = 0;
    while (in+1 < data.size()) {
        if (data[in].isSame(data[in+1])
            && data[in].getKind() == GHullFunction::AssignedToHull
            && data[in+1].getKind() == GHullFunction::AssignedToRace)
        {
            if (data[in].getPlayers().contains(data[in+1].getPlayers())) {
                /* in (the hull function) contains everything */
                data[out++] = data[in++];
                ++in;
            } else if (data[in+1].getPlayers().contains(data[in].getPlayers())) {
                /* in+1 (the racial ability) contains everything */
                ++in;
                data[out++] = data[in++];
            } else {
                /* no change, copy both */
                data[out++] = data[in++];
                data[out++] = data[in++];
            }
        } else {
            /* no match, copy one */
            data[out++] = data[in++];
        }
    }

    /* copy final item, if any */
    if (in < data.size())
        data[out++] = data[in++];
    
    data.resize(out);

    /* Other things we could do:
       - remove implied functions
       - combine level-restricted functions, i.e. if a ship has both a
         cloak-that-works-on-L1 and a cloak-that-works-on-L2, combine that
         into a cloak-that-works-on-L1-and-L2 */
}

/** Sort list for new ships. Brings it into a state that is useful for
    attaching it with a new (not-yet-built) ship. */
void
GHullFunctionList::sortForNewShip(GPlayerSet forPlayer)
{
    std::sort(data.begin(), data.end(), SortByPlayer(forPlayer));
}
