/**
  *  \file int/if/initif.cc
  *  \brief Interpreter Game Interface: Initialisation
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "int/exec.h"
#include "int/if/globalif.h"
#include "int/if/hullif.h"
#include "int/if/initif.h"
#include "int/if/ionif.h"
#include "int/if/iterif.h"
#include "int/if/mineif.h"
#include "int/if/planetif.h"
#include "int/if/playerif.h"
#include "int/if/shipif.h"
#include "int/if/specif.h"
#include "int/if/ufoif.h"
#include "int/if/vcrif.h"
#include "int/intglobal.h"
#include "util/modinit.h"
#include "int/if/msgif.h"
#include "int/if/drawingif.h"
#include "int/if/plugif.h"

static void
initInterpreterGameInterface()
{
    // Add global variables
    defineGlobalValue("BEAM",           new IntSimpleIndexableValue(IFBeamDim,      IFBeamGet,           IFBeamMake));
    defineGlobalValue("CADD",           new IntSimpleIndexableValue(0,              IFCAdd,              0));
    defineGlobalValue("CC$NOTIFYCONFIRMED", new IntCCNotifyConfirmed());            
    defineGlobalValue("CCOMPARE",       new IntSimpleIndexableValue(0,              IFCCompare,          0));
    defineGlobalValue("CDIV",           new IntSimpleIndexableValue(0,              IFCDiv,              0));
    defineGlobalValue("CEXTRACT",       new IntSimpleIndexableValue(0,              IFCExtract,          0));
    defineGlobalValue("CFG",            new IntSimpleIndexableValue(0,              IFCfgGet,            0));
    defineGlobalValue("CMUL",           new IntSimpleIndexableValue(0,              IFCMul,              0));
    defineGlobalValue("CREMOVE",        new IntSimpleIndexableValue(0,              IFCRemove,           0));
    defineGlobalValue("CSUB",           new IntSimpleIndexableValue(0,              IFCSub,              0));
    defineGlobalValue("DISTANCE",       new IntSimpleIndexableValue(0,              IFDistanceGet,       0));
    defineGlobalValue("ENGINE",         new IntSimpleIndexableValue(IFEngineDim,    IFEngineGet,         IFEngineMake));
    defineGlobalValue("FORMAT",         new IntSimpleIndexableValue(0,              IFFormat,            0));
    defineGlobalValue("HULL",           new IntSimpleIndexableValue(IFHullDim,      IFHullGet,           IFHullMake));
    defineGlobalValue("INMSG",          new IntSimpleIndexableValue(IFInmsgDim,     IFInmsgGet,          IFInmsgMake));
    defineGlobalValue("ISSPECIALFCODE", new IntSimpleIndexableValue(0,              IFIsSpecialFCodeGet, 0));
    defineGlobalValue("ITERATOR",       new IntSimpleIndexableValue(0,              IFIteratorGet,       0));
    defineGlobalValue("LAUNCHER",       new IntSimpleIndexableValue(IFTorpedoDim,   IFLauncherGet,       IFLauncherMake));
    defineGlobalValue("MARKER" ,        new IntSimpleIndexableValue(0,              0,                   IFDrawingMake));
    defineGlobalValue("MINEFIELD",      new IntSimpleIndexableValue(IFMinefieldDim, IFMinefieldGet,      IFMinefieldMake));
    defineGlobalValue("PLANET",         new IntSimpleIndexableValue(IFPlanetDim,    IFPlanetGet,         IFPlanetMake));
    defineGlobalValue("PLANETAT",       new IntSimpleIndexableValue(0,              IFPlanetAtGet,       0));
    defineGlobalValue("PLANETNAME",     new IntSimpleIndexableValue(IFPlanetDim,    IFPlanetNameGet,     0));
    defineGlobalValue("PLAYER",         new IntSimpleIndexableValue(IFPlayerDim,    IFPlayerGet,         IFPlayerMake));
    defineGlobalValue("PREF",           new IntSimpleIndexableValue(0,              IFPrefGet,           0));
    defineGlobalValue("RANDOM",         new IntSimpleIndexableValue(0,              IFRandom,            0));
    defineGlobalValue("RANDOMFCODE",    new IntSimpleIndexableValue(0,              IFRandomFCode,       0));
    defineGlobalValue("SHIP",           new IntSimpleIndexableValue(IFShipDim,      IFShipGet,           IFShipMake));
    defineGlobalValue("SHIPNAME",       new IntSimpleIndexableValue(IFShipDim,      IFShipNameGet,       0));
    defineGlobalValue("STORM",          new IntSimpleIndexableValue(IFIonDim,       IFIonGet,            IFIonMake));
    defineGlobalValue("SYSTEM.PLUGIN",  new IntSimpleIndexableValue(0,              IFSystemPlugin,      0));
    defineGlobalValue("TORPEDO",        new IntSimpleIndexableValue(IFTorpedoDim,   IFTorpedoGet,        IFTorpedoMake));
    defineGlobalValue("TRANSLATE",      new IntSimpleIndexableValue(0,              IFTranslate,         0));
    defineGlobalValue("TRUEHULL",       new IntSimpleIndexableValue(0,              IFTruehull,          0));
    defineGlobalValue("UFO",            new IntSimpleIndexableValue(IFUfoDim,       IFUfoGet,            IFUfoMake));
    defineGlobalValue("VCR",            new IntSimpleIndexableValue(IFVcrDim,       IFVcrGet,            IFVcrMake));

    defineGlobalProcedure("ADDCOMMAND",       IFAddCommand);
    defineGlobalProcedure("ADDCONFIG",        IFAddConfig);
    defineGlobalProcedure("ADDFCODE",         IFAddFCode);
    defineGlobalProcedure("AUTHPLAYER",       IFAuthPlayer);
    defineGlobalProcedure("CC$NOTIFY",        IFCCNotify);
    defineGlobalProcedure("CC$SELECTIONEXEC", IFCCSelectionExec);
    defineGlobalProcedure("NEWCIRCLE",        IFNewCircle);
    defineGlobalProcedure("NEWLINE",          IFNewLine);
    defineGlobalProcedure("NEWLINERAW",       IFNewLineRaw);
    defineGlobalProcedure("NEWMARKER",        IFNewMarker);
    defineGlobalProcedure("NEWRECTANGLE",     IFNewRectangle);
    defineGlobalProcedure("NEWRECTANGLERAW",  IFNewRectangleRaw);
    defineGlobalProcedure("SAVEGAME",         IFSaveGame);
    defineGlobalProcedure("SELECTIONLOAD",    IFSelectionLoad);
    defineGlobalProcedure("SELECTIONSAVE",    IFSelectionSave);

    // Add global context (=properties)
    IntExecutionContext::setNewGlobalContext(new IntGlobalContext());
}

// Constraints: after mod_int, with mod_intgame_gui or mod_intgame_console
const ModuleInitializer mod_intgame_common[] = {
    { OnStartup, initInterpreterGameInterface },
    { 0, 0 },
};
