/**
  *  \file reshack/colors.cc
  *  \brief Resource Editor: Colors
  *
  *  We can edit 8-bit and truecolor pictures. When we have a
  *  truecolor screen, all formats are supported; with 8-bit screen,
  *  only standard-palette and grayscale (=font) images are supported.
  *
  *  Pictures with less than 8 bits are expanded to 8 bits, truecolor
  *  pictures are converted to screen format (which is truecolor in
  *  this case).
  *
  *  Palettized images:
  *    The image palette is always current. For standard palette
  *    images, it is identical to the screen palette (which in turn is
  *    made identical to standard_palette).
  *
  *  Truecolor images:
  *    The images have no palette. Users deal with indices into
  *    standard_palette. Truecolor images always have the same format
  *    as the screen.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include "reshack/colors.h"
#include "cpluslib/assert.h"
#include "gfx/palette.h"
#include "ui/uiglobal.h"

/** Figure out color mode of an image. Note that this will never
    return GrayscaleColor, which is used internally only. */
ColorMode
identifyColorMode(Ptr<GfxPixmap> pix)
{
    if (pix->getBitsPerPixel() < 8) {
        /* palettized, but with a small palette */
        return ForeignPaletteColor;
    } else if (pix->getBitsPerPixel() == 8) {
        /* palettized */
        SDL_PixelFormat* ss = pix->getSystemSurface()->format;
        ASSERT(ss->palette != 0);
        if (ss->palette->ncolors != 256) {
            return ForeignPaletteColor;
        } else {
            bool same = true;
            for (int i = 0; i < COLOR_AVAIL; ++i) {
                if (ss->palette->colors[i].r != standard_palette[i].r
                    || ss->palette->colors[i].g != standard_palette[i].g
                    || ss->palette->colors[i].b != standard_palette[i].b)
                {
                    same = false;
                    break;
                }
            }
            if (same) {
                return StandardPaletteColor;
            } else {
                return ForeignPaletteColor;
            }
        }
    } else {
        /* truecolor */
        return TrueColor;
    }
}

/** Make picture editable. Converts the image to something we can deal with. */
Ptr<GfxPixmap>
makeEditable(Ptr<GfxPixmap> pix)
{
    /* We cannot edit transparency */
    pix->setColorKey(GFX_TRANSPARENT);

    /* Fix color mode */
    if (pix->getBitsPerPixel() < 8 || (pix->getBitsPerPixel() == 8 && pix->getSystemSurface()->format->palette->ncolors != 256)) {
        /* palettized with small palette. Increase to 8 bpp with full palette.
           Loaded images may have a less than full palette and will crash when
           we attempt to access a color beyond the original size. */
        Ptr<GfxPixmap> c = new GfxPixmap(pix->getWidth(), pix->getHeight(), 8, 0, 0, 0, 0);
        SDL_SetPalette(c->getSystemSurface(),
                       SDL_LOGPAL | SDL_PHYSPAL,
                       pix->getSystemSurface()->format->palette->colors,
                       0,
                       pix->getSystemSurface()->format->palette->ncolors);
        pix->blit(*c, 0, 0);
        return c;
    } else if (pix->getBitsPerPixel() == 8) {
        /* palettized, editable */
        return pix;
    } else {
        /* truecolor, convert to screen format */
        return pix->convertToScreenFormat();
    }
}

/** Check whether user is permitted to edit the specified color in a color mode.
    \param cm color mode
    \param color color number */
bool
isEditableColor(ColorMode cm, uint8_t color)
{
    switch (cm) {
     case StandardPaletteColor:
     case TrueColor:
     default:
        return color >= COLOR_AVAIL;

     case ForeignPaletteColor:
        return true;

     case GrayscaleColor:
        return false;
    }
}

/** Update screen palette to match image palette. Use for
    StandardPaletteColor images. */
void
setScreenPalette(Ptr<GfxPixmap> pix)
{
    // Pictures with more than 8bpp have no palette
    if (pix->getBitsPerPixel() > 8)
        return;

    // Remember palette
    for (int i = COLOR_AVAIL; i < NUM_STANDARD_COLORS; ++i) {
        standard_palette[i] = pix->getSystemSurface()->format->palette->colors[i];
    }

    // Copy palette
    if (ui_screen->getBitsPerPixel() <= 8) {
        // Copy raw palette
        SDL_SetPalette(ui_screen->getSystemSurface(),
                       SDL_LOGPAL | SDL_PHYSPAL,
                       &pix->getSystemSurface()->format->palette->colors[COLOR_AVAIL],
                       COLOR_AVAIL,
                       NUM_STANDARD_COLORS - COLOR_AVAIL);
        for (int i = COLOR_AVAIL; i < NUM_STANDARD_COLORS; ++i) {
            standard_colors[i] = i;
        }
    } else {
        // Update standard colors
        for (int i = COLOR_AVAIL; i < NUM_STANDARD_COLORS; ++i) {
            standard_colors[i] = SDL_MapRGB(ui_screen->getSystemSurface()->format,
                                            standard_palette[i].r,
                                            standard_palette[i].g,
                                            standard_palette[i].b);
        }
    }
}

/** Get screen color to use for a user color index.
    \param pix User pixmap
    \param index Color index */
uint32_t
colorIndexToScreenColor(Ptr<GfxPixmap> pix, uint8_t index)
{
    if (ui_screen->getBitsPerPixel() <= 8) {
        // Screen has 8 bpp or less, 1:1 map
        return index;
    } else if (pix->getBitsPerPixel() <= 8) {
        // Palettized image. Convert image palette to screen color
        return SDL_MapRGB(ui_screen->getSystemSurface()->format,
                          pix->getSystemSurface()->format->palette->colors[index].r,
                          pix->getSystemSurface()->format->palette->colors[index].g,
                          pix->getSystemSurface()->format->palette->colors[index].b);
    } else {
        // Truecolor screen with standard palette
        return standard_colors[index];
    }
}

/** Get image color to use for a user color index.
    \param pix User pixmap
    \param index Color index */
uint32_t
colorIndexToImageColor(Ptr<GfxPixmap> pix, uint8_t index)
{
    if (pix->getBitsPerPixel() <= 8) {
        // Image has 8 bpp or less, 1:1 map
        return index;
    } else {
        // Image has more than 8 bpp
        return SDL_MapRGB(pix->getSystemSurface()->format,
                          standard_palette[index].r,
                          standard_palette[index].g,
                          standard_palette[index].b);
    }
}
