/**
  *  \file resmgr/resmgr.cc
  *  \brief Resource Manager -- Interface Implementation
  *
  *  \todo A big problem of this system is lifetime management of the
  *  ResId strings. I definitely do not want a full string_t here for
  *  performance reasons; ResId should remain lightweight. However, we
  *  have the problem that getPixmap() etc clone the ResId and
  *  possibly keep it longer than the actual call. So far, all ResIds
  *  are constructed from static strings, but maybe that changes in
  *  the future. Even then, having users to think about this lifetime
  *  issue is bad. One possible solution would be to make a string
  *  pool which has the same lifetime as the respective cache.
  *
  *  (c) 2001-2022 Stefan Reuther <Streu@gmx.de>
  *
  *  This program is free software; you can redistribute it and/or
  *  modify it under the terms of file `COPYING' that comes with the
  *  source code.
  *
  *  This program is distributed in the hope that it will be useful,
  *  but WITHOUT ANY WARRANTY; without even the implied warranty of
  *  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.
  */

#include <cstring>
#include <algorithm>
#include "resmgr/resmgr.h"
#include "cpluslib/assert.h"
#include "util/nls.h"
#include "util/console.h"
#include "cpluslib/format.h"
#include "gfx/font.h"
#include "util/misc.h"
#include "util/random.h"
#include "util/except.h"
#include "io/textfile.h"

/********************************* ResId *********************************/

/** Generalize this resource Id. This modifies the resource Id
    in-place. */
void
ResId::generalize()
{
    if (id2)
        id2 = 0;
    else if (id1)
        id1 = 0;
}

/** Compare for equality. */
bool
ResId::operator==(const ResId& rhs) const
{
    return (name == rhs.name
            && id1 == rhs.id1
            && id2 == rhs.id2);
}

/** Compare less-than. */
bool
ResId::operator<(const ResId& rhs) const
{
    if (int i = name.compare(rhs.name))
        return i < 0;
    if (id1 < rhs.id1)
        return true;
    if (id1 > rhs.id1)
        return false;
    return id2 < rhs.id2;
}

/** Convert to string. This yields the external representation exposed
    to users. */
string_t
ResId::toString() const
{
    string_t rv = name;
    if (id1)
        rv += format("%d", id1);
    if (id2)
        rv += format("_%d", id2);
    return rv;
}

/** Check whether this resource Id matches the given template.
    \param name name to match (must match exactly)
    \param level level to match. 0=match ResId(name),
           1=match ResId(name,id1), 2=match ResId(name,id1,id2) */
bool
ResId::isElem(const char* name, int level) const
{
    int this_lvl = id2 ? 2 : id1 ? 1 : 0;
    return level == this_lvl
        && name == this->name;
}

/*************************** ResSimpleProvider ***************************/

/*! \class ResSimpleProvider

    This resource provider can be used as a template for implementing
    own resource providers. It does nothing, i.e., provides no
    resources. To make a simple resource provider, you can thus
    override only the methods you need. */
Ptr<GfxPixmap>
ResSimpleProvider::loadPixmap(ResId /*id*/)
{
    return 0;
}

SfxSampleData*
ResSimpleProvider::loadSound(ResId /*id*/)
{
    return 0;
}

Ptr<GfxAnimation>
ResSimpleProvider::loadAnimation(ResId /*id*/)
{
    return 0;
}

Ptr<GfxFont>
ResSimpleProvider::getFont(ResId /*id*/)
{
    return 0;
}

/************************* ResManager::SoundNode *************************/

ResManager::SoundNode::~SoundNode()
{ }

/******************************* ResManager ******************************/

/// Global resource manager instance.
ResManager ResManager::instance;

/*! \class ResManager

    This class manages a list of ResProvider's, and provides a cache
    for resource items that were already loaded.

    Resources are identified by ResId's. The ResManager automatically
    generalizes a resource Id if a detailed request cannot be
    satisfied.

    Resources are cached, and shared if possible. Inability to satisfy
    a request is cached as well, newly-added resource providers are
    not asked again. Therefore you should first initialize the
    ResManager using addProvider() before you ask it for data.

    Although you can create own instances of this class, you should
    use the prepared instance, see ResManager::getInstance(). */

/** Constructor. */
ResManager::ResManager()
{ }

/** Destructor. */
ResManager::~ResManager()
{
    clearPixmapCache();
    clearSoundCache();
}

/** Add a newly allocated provider. This assumes that the provider was
    just allocated with 'new', and it will take over ownership of the
    provider object. */
void
ResManager::addNewProvider(ResProvider* provider)
{
    provider_list.push_back(provider);
    clearPixmapCache();
    clearSoundCache();
}

/** Unload all resource providers of a given origin. */
void
ResManager::unloadByOrigin(const string_t& origin)
{
    size_t i = 0;
    bool did = false;
    while (i < provider_list.size()) {
        if (provider_list[i]->getOrigin() == origin) {
            provider_list.erase(provider_list.begin() + i);
            did = true;
        } else {
            ++i;
        }
    }

    if (did) {
        clearPixmapCache();
        clearSoundCache();
    }
}

/** Add pixmap associated with resource Id. Intended to be called by
    resource providers in response to their loadPixmap() call. */
void
ResManager::addPixmap(ResId id, Ptr<GfxPixmap> pix)
{
    if (pix) {
        pixmap_cache[id] = pix;
    }
}

/** Add sound associated with resource Id. Intended to be called by
    resource providers in response to their loadSound() call. */
void
ResManager::addSound(ResId id, SfxSampleData* snd)
{
    if (snd) {
        sound_cache_t::iterator i = sound_cache.find(id);
        if (i == sound_cache.end())
            i = safeInsert(sound_cache, id, new SoundNode);
        else if (i->second == 0)
            i->second = new SoundNode;

        i->second->sound = snd;
        // FIXME: what to do with the old one?
    }
}

/** Add animation associated with resource Id. Intended to be called
    by resource providers in response to their loadAnimation()
    call. */
void
ResManager::addAnimation(ResId id, Ptr<GfxAnimation> anim)
{
    if (anim) {
        anim_cache[id] = anim;
    }
}

/** Clear pixmap cache. This forgets all pixmaps currently known. */
void
ResManager::clearPixmapCache()
{
    pixmap_cache_t().swap(pixmap_cache);
}

/** Clear sound cache. This forgets all sounds currently known, and
    invalidates them.
    \see SfxSample::newGeneration() */
void
ResManager::clearSoundCache()
{
    for (sound_cache_t::iterator i = sound_cache.begin(); i != sound_cache.end(); ++i) {
        if (i->second && ! --i->second->use_count)
            delete i->second;
        i->second = 0;
    }
    sound_cache_t().swap(sound_cache);
    SfxSample::newGeneration();
}

/** Clear animation cache. This forgets all animations currently known. */
void
ResManager::clearAnimationCache()
{
    anim_cache_t().swap(anim_cache);
}

/** Get pixmap. If there is a pixmap associated with the resource Id,
    return it. If there are multiple pixmaps, return an arbitrarily
    chosen one. If there is no pixmap, returns 0. */
Ptr<GfxPixmap>
ResManager::getPixmap(ResId id)
{
    return getPixmapNode(id);
}

/** Get pixmap. Like getPixmap(), but never returns null. Throws
    FileOpenException if image not found. */
Ptr<GfxPixmap>
ResManager::getPixmapT(ResId id)
{
    Ptr<GfxPixmap> p = getPixmap(id);
    if (!p)
        throw FileOpenException("g." + id.toString(), _("Resource not found"));
    return p;
}

/** Get sound. If there is a sound associated with the resource Id,
    return it. If there are multiple sounds, return an arbitrarily
    chosen one. If there is no sound, returns 0. */
SfxSample
ResManager::getSound(ResId id)
{
    // We cannot load any sound when we don't have sound output; we
    // don't know where to convert it to.
    if (!isSoundAvailable())
        return 0;

    SoundNode* n = getSoundNode(id);
    if (n) {
        return n->sound;
    } else {
        return 0;
    }
}

/** Get animation. If there is an animation associated with the resource Id,
    return it. If there are multiple animations, return an arbitrarily
    chosen one. If there is no animation, returns 0. */
Ptr<GfxAnimation>
ResManager::getAnimation(ResId id)
{
    return getAnimationNode(id);
}

/** Get pixmap node for resource Id. First, attempt to load relevant
    data from resource managers. If this fails, attempt to load
    generalized data. */
Ptr<GfxPixmap>
ResManager::getPixmapNode(ResId id)
{
    /* already in cache? */
    pixmap_cache_t::iterator i = pixmap_cache.find(id);
    if (i != pixmap_cache.end())
        return i->second;

    /* ask resource providers */
    i = pixmap_cache.insert(pixmap_cache_t::value_type(id, 0)).first;
    for (uint32_t j = provider_list.size(); j > 0; --j) {
        try {
            if (Ptr<GfxPixmap> pix = provider_list[j-1]->loadPixmap(id)) {
                i->second = pix;
                break;
            }
        }
        catch (...)
        { }
    }
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        Ptr<GfxPixmap> n = getPixmapNode(g);
        if (n && !i->second) {
            i->second = n;
        }
        return i->second;
    }
    return 0;
}

/** Get sound node for resource Id. First, attempt to load relevant
    data from resource managers. If this fails, attempt to load
    generalized data. */
ResManager::SoundNode*
ResManager::getSoundNode(ResId id)
{
    /* already in cache? */
    sound_cache_t::iterator i = sound_cache.find(id);
    if (i != sound_cache.end())
        return i->second;

    /* ask resource providers */
    i = sound_cache.insert(sound_cache_t::value_type(id, 0)).first;
    for (uint32_t j = provider_list.size(); j > 0; --j) {
        try {
            if (SfxSampleData* sfx = provider_list[j-1]->loadSound(id)) {
                addSound(id, sfx);
                break;
            }
        }
        catch (...)
        { }
    }
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        SoundNode* n = getSoundNode(g);
        if (n && !i->second) {
            i->second = n;
            ++n->use_count;
        }
        return i->second;
    }
    return 0;
}

/** Get animation node for resource Id. First, attempt to load
    relevant data from resource managers. If this fails, attempt to
    load generalized data. */
Ptr<GfxAnimation>
ResManager::getAnimationNode(ResId id)
{
    /* already in cache? */
    anim_cache_t::iterator i = anim_cache.find(id);
    if (i != anim_cache.end())
        return i->second;

    /* ask resource providers */
    i = anim_cache.insert(anim_cache_t::value_type(id, 0)).first;
    for (uint32_t j = provider_list.size(); j > 0; --j) {
        try {
            if (Ptr<GfxAnimation> ani = provider_list[j-1]->loadAnimation(id)) {
                i->second = ani;
                break;
            }
        }
        catch (...)
        { }
    }
    if (i->second)
        return i->second;

    /* still not found? try to generalize if possible */
    if (!id.isGeneric()) {
        ResId g = id;
        g.generalize();
        Ptr<GfxAnimation> n = getAnimationNode(g);
        if (n && !i->second) {
            i->second = n;
        }
        return i->second;
    }
    return 0;
}

/** Get font. Streams are not cached in any way, we just ask all the
    providers whether they know it. */
Ptr<GfxFont>
ResManager::getFont(ResId id)
{
    for (uint32_t j = provider_list.size(); j > 0; --j)
        if (Ptr<GfxFont> r = provider_list[j-1]->getFont(id))
            return r;
    return 0;
}

/**************************** openResourceFile ***************************/

const char* const gfx_suffix_list[] = { "cd", "gfx", "bmp", "png", "jpg", 0 };
const char* const sfx_suffix_list[] = { "wav", 0 };
const char* const fnt_suffix_list[] = { "fnt", 0 };
const char* const null_suffix_list[] = { 0 };

/** Open a resource file. If the specified file name ends with a dot, this
    searches for a file according to the suffix list. Otherwise, only the
    exact name specified is attempted.

    \param dir      Directory to look in
    \param filename (User-)specified file name
    \param suffixes suffix list, null-terminated
    \return stream, or null */
Ptr<Stream>
openResourceFile(Directory& dir, const string_t filename, const char*const suffixes[])
{
    if (filename.size() != 0 && filename[filename.size()-1] == '.') {
        /* This means: look with suffixes */
        for (int i = 0; suffixes[i] != 0; ++i) {
            const char* p = suffixes[i];
            if (*p == '.')
                ++p;
            if (Ptr<Stream> s = dir.openFileNT(filename + p, Stream::C_READ))
                return s;
        }
    }

    /* Nothing found or no suffix search requested: try opening as-is */
    return dir.openFileNT(filename, Stream::C_READ);
}
